import time
import shlex, subprocess
import opts

def run_cmd(cmd, bg=False):
  args = shlex.split(cmd)
  start = int(round(time.time() * 1000))
  ret = ""
  if not bg:
    ret = subprocess.check_output(args)
  else:
    subprocess.Popen(args)

  end = int(round(time.time() * 1000))
  if opts.VERBOSE and end - start > 50:
    print "TIME:", end - start, "CMD", cmd.split()[0]
  return ret




#!/usr/bin/env python
# required utilities: imagemagick, xdotool, xbacklight

import time
import os
import math

try: import cpickle as pickle
except: import pickle

import opts
import models

from run_cmd import run_cmd
def set_brightness(new_level, time):
  if opts.XRANDR_OUTPUT:
    import xrandr
    xrandr.set_brightness(new_level, time)
  elif opts.LIGHT_OUTPUT:
    import light
    light.set_brightness(new_level, time)
  else:
    import xbacklight
    xbacklight.set_brightness(new_level, time)

def get_brightness():
  if opts.XRANDR_OUTPUT:
    import xrandr
    return xrandr.get_brightness()
  elif opts.LIGHT_OUTPUT:
    import light
    return light.get_brightness()
  else:
    import xbacklight
    return xbacklight.get_brightness()

def get_hour():
  hour = int(time.strftime("%H")) * 60
  hour_slice = round(int(time.strftime("%M")) / models.HOUR_SLICE) * models.HOUR_SLICE
  hour = hour + hour_slice
  return hour

def fmt_hour(minutes):
  hour = minutes / 60
  mnt = int((minutes % 60) / models.HOUR_SLICE) * models.HOUR_SLICE

  return "%02i:%02i" % (hour, mnt)


PREV_LEVELS={}
PREV_WINDOWS = []
MAX_WINDOWS=100

def add_prev_level(window, new_level):
  global PREV_LEVELS
  PREV_LEVELS = {}
  PREV_WINDOWS.append((window, new_level))
  while len(PREV_WINDOWS) > MAX_WINDOWS:
    PREV_WINDOWS.pop(0)

  for datum in PREV_WINDOWS:
    win, level = datum
    PREV_LEVELS[win] = max(level, 1)

def get_window():
  focused_cmd = opts.CHECK_TITLE_CMD
  if opts.CHECK_PID:
    focused_cmd = opts.CHECK_PID_CMD

  try: window = run_cmd(focused_cmd).strip()
  except Exception, e: print e; window = None

  return window

def monitor_luma():
  prev_brightness = None
  prev_window = None
  prev_mean = None
  faded = None

  cur_range = opts.MAX_WHITE - opts.MIN_WHITE
  suppressed_time = 0

  last_screenshot = int(time.time())


  while True:
    time.sleep(opts.SLEEP_TIME / 1000.0)

    window = get_window()

    now = time.time()
    if prev_window != window:
      if not faded:
        if window in PREV_LEVELS:
          pred = PREV_LEVELS[window]
          fade = opts.TRANSITION_MS / 2
          if opts.VERBOSE:
            curtime = int(time.time())
            print "PRIOR|TS:%s," % curtime, "RECALLED BRIGHTNESS:", "%s/%s" % (int(pred), MAX_LEVEL), "FOR", window[:15]

          set_brightness(pred, fade)
          faded = True

          continue

    if now - last_screenshot < opts.SCREENSHOT_TIME / 1000.0:
      continue

    last_screenshot = now
    faded = False



    if prev_window == window:
      suppressed_time += opts.SCREENSHOT_TIME / 1000

      if opts.LEARN_MODE and prev_brightness:
        models.add_observation(prev_brightness, prev_mean)

      if opts.RECALIBRATE_MS > 0 and suppressed_time < opts.RECALIBRATE_MS:
          continue

      print "RECALIBRATING BRIGHTNESS AFTER %S ms" % opts.RECALIBRATE_MS

    suppressed_time = 0

    window = get_window()
    prev_window = window

    brightness = run_cmd(opts.SCREENSHOT_CMD + " " + opts.BRIGHTNESS_CMD)


    try:
      cur_mean = float(brightness)
    except Exception, e:
      print "ERROR GETTING MEAN LUMA", e
      continue


    trimmed_mean = max(min(opts.MAX_WHITE, cur_mean), opts.MIN_WHITE) - opts.MIN_WHITE
    trimmed_mean = int(trimmed_mean / models.LUMA_BUCKET) * models.LUMA_BUCKET
    range_is = float(trimmed_mean) / float(cur_range)

    new_gamma = 1 - range_is
    hour = get_hour()
    new_level =  (opts.MAX_LEVEL - opts.MIN_LEVEL) * new_gamma + opts.MIN_LEVEL

    pred_level = models.get_mean_brightness(hour, trimmed_mean)
    if pred_level is not None:
      new_level = pred_level



    prev_mean = trimmed_mean

    new_level = max(round(new_level), 1)
    if prev_brightness != new_level:
      now = int(time.time())
      print "MODEL|TS:%s," % now, "LUMA:%05i," % trimmed_mean, "NEW GAMMA:%.02f," % new_gamma, "NEW BRIGHTNESS:", "%s/%s" % (int(new_level), opts.MAX_LEVEL)

      set_brightness(new_level, opts.TRANSITION_MS / 2)

      add_prev_level(window, new_level)
    prev_brightness = new_level

def run():
  opts.load_options()

  if opts.RESET:
    set_brightness(100, 1)
  elif opts.ADJUSTMENT:
    b = get_brightness()
    nb = b + opts.ADJUSTMENT
    set_brightness(nb, 0)
  elif opts.VIZ_LUMA_MAP:
    import luma_vis
    luma_vis.visualize(models.LUMA_FILE)
  else:
    opts.print_config()
    models.load_luma_observations()
    models.load_luma_map(models.LUMA_FILE)

    if opts.RUN_AS_DAEMON:
        import os
        try:
            import daemon
            import daemon.pidfile
        except:
            print "MISSING DAEMON MODULE. PLEASE INSTALL PYTHON-DAEMON TO ENABLE DAEMON MODE"
            import sys
            sys.exit(1)

        uid = os.getuid()
        lock_file = os.path.join(os.path.sep, "tmp", "autolux.%s.pid" % uid)
        print "RUNNING IN DAEMON MODE"
        print "LOCKFILE", lock_file
        with daemon.DaemonContext(pidfile=daemon.pidfile.PIDLockFile(lock_file)):
            monitor_luma()
    else:
        monitor_luma()

if __name__ == "__main__":
  try:
    run()
  except:
    # we always reset brightness back to 100% for xrandr
    # if there was an error
    if opts.XRANDR_OUTPUT:
      set_brightness(100, 1)



import os
import time

import models

# BRIGHTNESS LEVELS (should be between 1 and 100)
MIN_LEVEL=5
MAX_LEVEL=100

# interpolate over our threshold (should be between 1 and 65K)
MAX_WHITE=60000
MIN_WHITE=5000

# interval between screenshots
SLEEP_TIME=67
SCREENSHOT_TIME=1200
TRANSITION_MS=800
RECALIBRATE_MS=60 * 1000

# EXAMPLE: 100x200+300+400
# 100 width, 200 height, 300 offset from left, 400 offset from top
CROP_SCREEN="10x100%+400+0"
HORIZ_CROP_SCREEN="90%x10+200+400"

SCREENSHOT_CMD='import -silent -colorspace gray -screen -w root -quality 20'
BRIGHTNESS_CMD='-format "%[mean]" info:'

# change brightness when PID changes or
# change brightness when window changes
CHECK_PID=False
CHECK_PID_CMD='xdotool getwindowfocus getwindowpid'
# change brightness when window name changes
CHECK_TITLE_CMD='xdotool getwindowfocus getwindowname'

# default to True, now that we can skip using xbacklight
LEARN_MODE=True
VIZ_LUMA_MAP=False

PLOT_LUMA=True
PLOT_BRIGHT=False

RUN_AS_DAEMON=False

# do we use software dimming or not
XRANDR_OUTPUT = None
ADJUSTMENT = None
RESET = False

# do we use `light` to adjust the brightness
LIGHT_OUTPUT = None

VERBOSE=False
def load_options():
  global MIN_LEVEL, MAX_LEVEL, MAX_WHITE, MIN_WHITE, CROP_SCREEN
  global SLEEP_TIME, TRANSITION_MS, RECALIBRATE_MS, SCREENSHOT_TIME
  global VERBOSE, CHECK_PID, LEARN_MODE,VIZ_LUMA_MAP
  global PLOT_LUMA, PLOT_BRIGHT
  global RUN_AS_DAEMON, XRANDR_OUTPUT
  global ADJUSTMENT, RESET
  global LIGHT_OUTPUT

  from optparse import OptionParser
  parser = OptionParser()
  parser.add_option("--daemon", dest="run_as_daemon", help="run autolux as a daemon",
    default=RUN_AS_DAEMON, action="store_true")
  parser.add_option("--file", dest="luma_file", help="luma file to load", default=models.LUMA_FILE_DEFAULT)
  parser.add_option("--sleep-interval", dest="sleep_interval", type="int", default=SLEEP_TIME,
    help="check for window change ever SLEEP_INTERVAL ms, default is %s" % SLEEP_TIME)
  parser.add_option("--interval", dest="interval", type="int", default=SCREENSHOT_TIME,
    help="take screen snapshot every INTERVAL ms and readjust the screen brightness, default is %s" % SCREENSHOT_TIME)
  parser.add_option("--min", "--min-level", dest="min_level", type="int", default=MIN_LEVEL,
    help="min brightness level (from 1 to 100, default is %s)" % MIN_LEVEL)
  parser.add_option("--max", "--max-level", dest="max_level", type="int", default=MAX_LEVEL,
    help="max brightness level (from 1 to 100, default is %s)" % MAX_LEVEL)
  parser.add_option("--lower", "--lower-threshold", dest="min_white", type="int", default=MIN_WHITE,
    help="lower whiteness threshold before setting screen to highest brightness (1K to 15K, default is %s)" % MIN_WHITE)
  parser.add_option("--upper", "--upper-threshold", dest="max_white", type="int", default=MAX_WHITE,
    help="upper whiteness threshold before setting screen to lowest brightness (45K to 65K, default is %s)" % MAX_WHITE)
  parser.add_option("--recalibrate-time", dest="recalibrate", type="int",
    default=RECALIBRATE_MS, help="ms before recalibrating even if the window hasn't changed. set to 0 to disable, default is 60K")
  parser.add_option("--fade-time", dest="fade_time", type="int", default=TRANSITION_MS,
    help="time to fade backlight in ms, default is %s" % TRANSITION_MS)
  parser.add_option("--crop", dest="crop_screen", type='str', default=CROP_SCREEN,
    help="area to inspect, use imagemagick geometry style string (f.e. 50%x20%+400+100 means 50% width, 20% height at offset 400x and 100y)")
  parser.add_option("--pid", dest="check_pid", action="store_true", help="check screen brightness when PID changes")
  parser.add_option("--title", dest="check_pid", action="store_false", help="check screen brightness when window changes")
  parser.add_option("--horizontal", dest="horizontal", action="store_true", help="take a horizontal screenshot instead of vertical")
  parser.add_option("--no-learn", dest="learn", action="store_false", help="disable learning", default=LEARN_MODE)
  parser.add_option("--verbose", dest="verbose", action="store_true", help="turn on verbose output, including screenshot timing info")
  parser.add_option("--visualize", dest="visualize", action="store_true", help="visualize your brightness model", default=VIZ_LUMA_MAP)
  parser.add_option("--plot-luma", dest="plot_luma", action="store_true", help="plot screen luminence on y axis and predicted brightness as color, good for observing prefered brightness by time of day", default=PLOT_LUMA)
  parser.add_option("--plot-brightness", dest="plot_luma", action="store_false", help="plot predicted brightness on y axis and input luminence as color, good for observing eye strain", default=not PLOT_LUMA)

  parser.add_option("--xrandr", dest="xrandr_output", type="str", default=None)

  parser.add_option("--adjust", dest="adjustment", type="float", default=None)
  parser.add_option("--reset", dest="reset", action="store_true", default=None)
  parser.add_option("--use-light", dest="light_output", action="store_true")



  options, args = parser.parse_args()
  MIN_LEVEL = options.min_level
  MAX_LEVEL = options.max_level
  RUN_AS_DAEMON = options.run_as_daemon
  SCREENSHOT_TIME = options.interval
  SLEEP_TIME = options.sleep_interval
  MAX_LEVEL = options.max_level
  TRANSITION_MS = options.fade_time
  CROP_SCREEN = options.crop_screen
  VERBOSE = options.verbose
  RECALIBRATE_MS = options.recalibrate
  CHECK_PID = options.check_pid
  LEARN_MODE=options.learn
  VIZ_LUMA_MAP=options.visualize
  models.LUMA_FILE=options.luma_file
  PLOT_BRIGHT=not options.plot_luma
  PLOT_LUMA=options.plot_luma
  XRANDR_OUTPUT=options.xrandr_output
  ADJUSTMENT=options.adjustment
  RESET=options.reset
  LIGHT_OUTPUT=options.light_output

  MIN_WHITE = options.min_white
  MAX_WHITE = options.max_white


  if options.horizontal:
    CROP_SCREEN = HORIZ_CROP_SCREEN

  global SCREENSHOT_CMD
  if CROP_SCREEN is not None:
    SCREENSHOT_CMD += ' -crop %s' % CROP_SCREEN




def print_config():
  print "DAEMON MODE:", not not RUN_AS_DAEMON
  print "CROPPING:", not not CROP_SCREEN
  print "FADE TIME:", TRANSITION_MS
  print "SLEEP TIME:", SCREENSHOT_TIME
  print "DISPLAY RANGE:", MIN_LEVEL, MAX_LEVEL
  print "LEARNING MODE:", LEARN_MODE
  print "BRIGHTNESS RANGE:", MIN_WHITE, MAX_WHITE
  print "RECALIBRATE EVERY:", RECALIBRATE_MS
  print "FOLLOW WINDOW PID:", not not CHECK_PID
  print "FOLLOW WINDOW TITLE:", not CHECK_PID
  print "SCREENSHOT CMD", SCREENSHOT_CMD





# A small compatibility layer for `light`
# https://github.com/haikarainen/light

import opts
from run_cmd import run_cmd

def set_brightness(new_level, time):
    run_cmd("light -S {:f}".format(new_level))

def get_brightness():
    out = run_cmd("light -G")
    return float(out.strip())



# a compatibility layer for xrandr meant to set brightness
import opts
from run_cmd import run_cmd

def set_brightness(new_level, time):
    output = opts.XRANDR_OUTPUT

    # calibrating to xbacklight 1 - 100 by dividing by 100
    new_level /= 100.0

    new_level = min(new_level, 1.0)
    new_level = max(new_level, 0.3)

    run_cmd("xrandr --output %s --brightness %s" % (output, new_level))

def get_brightness():
    out = run_cmd("xrandr --current --verbose")
    for line in out.split("\n"):
        if line.find("Brightness") != -1:
            brt = line.split(":")[1]
            brt = float(brt)
    # calibrating to xbacklight 1 - 100 by multiplying by 100
    return brt * 100
